from regressions.errors.analyses import select_equations
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import positive_integer
from regressions.statistics.rounding import rounded_value
from .maxima import maxima_points
from .minima import minima_points
from .intervals import sign_chart

def extrema_points(equation_type, coefficients, precision = 4):
    """
    Calculates the extrema of a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which extrema must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points['maxima'] : list of float or str
        Values of the x-coordinates at which the original function has a relative maximum; if the function is sinusoidal, then only two or three results within a two-period interval will be listed, but a general form will also be included; if the function has no maxima, then it will return a list of `None`
    points['minima'] : list of float or str
        Values of the x-coordinates at which the original function has a relative minimum; if the function is sinusoidal, then only two or three results within a two-period interval will be listed, but a general form will also be included; if the function has no minima, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.criticals.critical_points`, :func:`~regressions.analyses.intervals.sign_chart`, :func:`~regressions.analyses.maxima.maxima_points`, :func:`~regressions.analyses.minima.minima_points`, :func:`~regressions.analyses.points.key_coordinates`

    Notes
    -----
    - Critical points for the derivative of a function: :math:`c_i = \\{ c_1, c_2, c_3,  \\cdots, c_{n-1}, c_n \\}`
    - X-coordinates of the extrema of the function: :math:`x_{ext} = \\{ x \\mid x \\in c_i, \\left( f'(\\frac{c_{j-1} + c_j}{2}) < 0 \\cap f'(\\frac{c_j + c_{j+1}}{2}) > 0 \\right) \\\\ \\cup \\left( f'(\\frac{c_{j-1} + c_j}{2}) > 0 \\cap f'(\\frac{c_j + c_{j+1}}{2}) < 0 \\right) \\}`
    - |extrema_values|

    Examples
    --------
    Import `extrema_points` function from `regressions` library
        >>> from regressions.analyses.extrema import extrema_points
    Calulate the extrema of a cubic function with coefficients 1, -15, 63, and -7
        >>> points_cubic = extrema_points('cubic', [1, -15, 63, -7])
        >>> print(points_cubic['maxima'])
        [3.0]
        >>> print(points_cubic['minima'])
        [7.0]
    Calulate the extrema of a sinusoidal function with coefficients 2, 3, 5, and 7
        >>> points_sinusoidal = extrema_points('sinusoidal', [2, 3, 5, 7])
        >>> print(points_sinusoidal['maxima'])
        [5.5236, 7.618, 9.7124, '5.5236 + 2.0944k']
        >>> print(points_sinusoidal['minima'])
        [6.5708, 8.6652, '6.5708 + 2.0944k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    positive_integer(precision)
    
    # Determine maxima and minima
    max_points = maxima_points(equation_type, coefficients, precision)
    min_points = minima_points(equation_type, coefficients, precision)
    
    # Create dictionary to return
    result = {}

    # Handle sinusoidal case
    if equation_type == 'sinusoidal':
        # Recreate sign chart
        intervals_set = sign_chart('sinusoidal', coefficients, 1, precision)

        # Grab general form
        general_form = intervals_set[-1]

        # Extract periodic unit
        periodic_unit_index = general_form.find(' + ') + 3
        periodic_unit = 2 * float(general_form[periodic_unit_index:-1])
        rounded_periodic_unit = rounded_value(periodic_unit, precision)

        # Create general forms for max and min
        max_general_form = str(max_points[0]) + ' + ' + str(rounded_periodic_unit) + 'k'
        min_general_form = str(min_points[0]) + ' + ' + str(rounded_periodic_unit) + 'k'

        # Append general form as final element of each list
        max_extended = max_points + [max_general_form]
        min_extended = min_points + [min_general_form]
        result = {
            'maxima': max_extended,
            'minima': min_extended
        }
    
    # Handle all other cases
    else:
        result = {
            'maxima': max_points,
            'minima': min_points
        }
    return result


from regressions.errors.matrices import square_matrix
from .multiplication import scalar_product_matrix
from .determinant import linear_determinant
from .transpose import transposed_matrix
from .minors import matrix_of_minors
from .cofactors import matrix_of_cofactors

def inverse_matrix(matrix):
    """
    Generate the inverse matrix of a given matrix

    Parameters
    ----------
    matrix : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within first argument must be integers or floats
    ValueError
        First argument must contain the same amount of lists as the amount of elements contained within its first list
    
    Returns
    -------
    inverse : list of lists of float
        List of lists corresponding to the inverse of the original matrix; if original matrix has a determinant of zero, then 0.0001 will be used as its determinant, ensuring a result

    See Also
    --------
    :func:`~regressions.matrices.cofactors.matrix_of_cofactors`, :func:`~regressions.matrices.minors.matrix_of_minors`, :func:`~regressions.matrices.transpose.transposed_matrix`, :func:`~regressions.matrices.determinant.linear_determinant`, :func:`~regressions.matrices.multiplication.scalar_product_matrix`, :func:`~regressions.matrices.solve.system_solution`

    Notes
    -----
    - Original matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Inverse of matrix: :math:`\\mathbf{A}^{-1} = \\frac{1}{|\\mathbf{A}|}\\cdot{{{\\mathbf{A}^M}^C}^T}`
    - |inverse|

    Examples
    --------
    Import `inverse_matrix` function from `regressions` library
        >>> from regressions.matrices.inverse import inverse_matrix
    Generate the inverse of [[1, 2], [3, 4]]
        >>> inverse_2x2 = inverse_matrix([[1, 2], [3, 4]])
        >>> print(inverse_2x2)
        [[-2.0, 1.0], [1.5, -0.5]]
    Generate the inverse of [[2, 3, 5], [7, 11, 13], [17, 19, 23]]
        >>> inverse_3x3 = inverse_matrix([[2, 3, 5], [7, 11, 13], [17, 19, 23]])
        >>> print(inverse_3x3)
        [[-0.07692307692307693, -0.3333333333333333, 0.20512820512820512], [-0.7692307692307692, 0.5, -0.11538461538461538], [0.6923076923076923, -0.16666666666666666, -0.01282051282051282]]
    """
    # Handle input errors
    square_matrix(matrix)

    # Determine determinant of matrix
    determinant = linear_determinant(matrix)

    # Circumvent division by zero
    if determinant == 0:
        determinant = 0.0001
    
    # Calculate reciprocal of determinant
    determinant_reciprocal = 1 / determinant
    
    # Create adjugate of original matrix by transposing its cofactors
    adjugate = transposed_matrix(matrix_of_cofactors(matrix_of_minors(matrix)))

    # Generate inverse matrix by multiplying adjugate by reciprocal
    result = scalar_product_matrix(adjugate, determinant_reciprocal)
    return result


from regressions.errors.scalars import four_scalars, five_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.sort import sorted_list
from regressions.statistics.rounding import rounded_value, rounded_list
from regressions.analyses.derivatives.cubic import cubic_derivatives
from .quadratic import quadratic_roots
from .linear import linear_roots

def cubic_roots(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    """
    Calculates the roots of a cubic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the cubic term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the quadratic term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Coefficient of the linear term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    fourth_constant : int or float
        Coefficient of the constant term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First four arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    roots : list of float
        List of the x-coordinates of all of the x-intercepts of the original function

    See Also
    --------
    :func:`~regressions.analyses.equations.cubic.cubic_equation`, :func:`~regressions.analyses.derivatives.cubic.cubic_derivatives`, :func:`~regressions.analyses.integrals.cubic.cubic_integral`, :func:`~regressions.models.cubic.cubic_model`

    Notes
    -----
    - Standard form of a cubic function: :math:`f(x) = a\\cdot{x^3} + b\\cdot{x^2} + c\\cdot{x} + d`
    - Cubic formula: :math:`x_k = -\\frac{1}{3a}\\cdot(b + \\xi^k\\cdot{\\eta} + \\frac{\\Delta_0}{\\xi^k\\cdot{\\eta}})`

        - :math:`\\Delta_0 = b^2 - 3ac`
        - :math:`\\Delta_1 = 2b^3 - 9abc +27a^2d`
        - :math:`\\xi = \\frac{-1 + \\sqrt{-3}}{2}`
        - :math:`\\eta = \\sqrt[3]{\\frac{\\Delta_1 \\pm \\sqrt{\\Delta_1^2 - 4\\Delta_0^3}}{2}}`
        - :math:`k \\in \\{ 0, 1, 2 \\}`
    
    - |cubic_formula|

    Examples
    --------
    Import `cubic_roots` function from `regressions` library
        >>> from regressions.analyses.roots.cubic import cubic_roots
    Calculate the roots of a cubic function with coefficients 2, 3, 5, and 7
        >>> roots_first = cubic_roots(2, 3, 5, 7)
        >>> print(roots_first)
        [-1.4455]
    Calculate the roots of a cubic function with coefficients 7, -5, -3, and 2
        >>> roots_second = cubic_roots(7, -5, -3, 2)
        >>> print(roots_second)
        [-0.6431, 0.551, 0.8064]
    Calculate the roots of a cubic function with all inputs set to 0
        >>> roots_zeroes = cubic_roots(0, 0, 0, 0)
        >>> print(roots_zeroes)
        [-1.0]
    """
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant, fourth_constant], precision)

    # Create intermediary variables
    xi = (-1 + (-3)**(1/2)) / 2
    delta_first = coefficients[1]**2 - 3 * coefficients[0] * coefficients[2]
    delta_second = 2 * coefficients[1]**3 - 9 * coefficients[0] * coefficients[1] * coefficients[2] + 27 * coefficients[0]**2 * coefficients[3]
    discriminant = delta_second**2 - 4 * delta_first**3
    eta_first = ((delta_second + discriminant**(1/2)) / 2)**(1/3)
    eta_second = ((delta_second - discriminant**(1/2)) / 2)**(1/3)
    eta = 0
    if eta_first == 0:
        eta = eta_second
    else:
        eta = eta_first
    
    # Create roots
    roots = []
    first_root = (-1 / (3 * coefficients[0])) * (coefficients[1] + eta * xi**0 + delta_first / (eta * xi**0))
    second_root = (-1 / (3 * coefficients[0])) * (coefficients[1] + eta * xi**1 + delta_first / (eta * xi**1))
    third_root = (-1 / (3 * coefficients[0])) * (coefficients[1] + eta * xi**2 + delta_first / (eta * xi**2))

    # Identify real and imaginary components of complex roots
    first_real = first_root.real
    second_real = second_root.real
    third_real = third_root.real
    first_imag = first_root.imag
    second_imag = second_root.imag
    third_imag = third_root.imag

    # Determine magnitudes of imaginary components
    size_first_imag = (first_imag**2)**(1/2)
    size_second_imag = (second_imag**2)**(1/2)
    size_third_imag = (third_imag**2)**(1/2)

    # Eliminate roots with large imaginary components
    if size_first_imag < 10**(-precision):
        first_root = first_real
        roots.append(first_root)
    if size_second_imag < 10**(-precision):
        second_root = second_real
        roots.append(second_root)
    if size_third_imag < 10**(-precision):
        third_root = third_real
        roots.append(third_root)
    
    # Eliminate duplicate roots
    unique_roots = list(set(roots))
    
    # Sort unique roots
    sorted_roots = sorted_list(unique_roots)

    # Round roots
    rounded_roots = rounded_list(sorted_roots, precision)

    # Return result
    result = rounded_roots
    return result

def cubic_roots_first_derivative(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)

    # Generate coefficients of first derivative
    constants = cubic_derivatives(first_constant, second_constant, third_constant, fourth_constant)['first']['constants']

    # Determine roots of first derivative
    result = quadratic_roots(*constants, precision)
    return result

def cubic_roots_second_derivative(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)

    # Generate coefficients of second derivative
    constants = cubic_derivatives(first_constant, second_constant, third_constant, fourth_constant)['second']['constants']

    # Determine roots of second derivative
    result = linear_roots(*constants, precision)
    return result

def cubic_roots_initial_value(first_constant, second_constant, third_constant, fourth_constant, initial_value, precision = 4):
    # Handle input errors
    five_scalars(first_constant, second_constant, third_constant, fourth_constant, initial_value)
    positive_integer(precision)

    # Determine roots given an initial value
    result = cubic_roots(first_constant, second_constant, third_constant, fourth_constant - initial_value, precision)
    return result

def cubic_roots_derivative_initial_value(first_constant, second_constant, third_constant, fourth_constant, initial_value, precision = 4):
    # Handle input errors
    five_scalars(first_constant, second_constant, third_constant, fourth_constant, initial_value)
    positive_integer(precision)

    # Determine roots of derivative given an initial value
    result = quadratic_roots(3 * first_constant, 2 * second_constant, third_constant - initial_value, precision)
    return result


from regressions.errors.analyses import select_equations
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import positive_integer
from .roots.linear import linear_roots
from .roots.quadratic import quadratic_roots
from .roots.cubic import cubic_roots
from .roots.hyperbolic import hyperbolic_roots
from .roots.exponential import exponential_roots
from .roots.logarithmic import logarithmic_roots
from .roots.logistic import logistic_roots
from .roots.sinusoidal import sinusoidal_roots

def intercept_points(equation_type, coefficients, precision = 4):
    """
    Calculates the roots of a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which intercepts must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points : list of float or str
        Values of the x-coordinates at which the original function crosses the x-axis; if the function is sinusoidal, then only the initial results within a four-period interval will be listed, but general forms will also be included; if the function has no x-intercepts, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.criticals.critical_points`, :func:`~regressions.analyses.intervals.sign_chart`, :func:`~regressions.analyses.points.key_coordinates`

    Notes
    -----
    - Domain of a function: :math:`x_i = \\{ x_1, x_2, \\cdots, x_n \\}`
    - X-intercepts (roots) of the function: :math:`x_r = \\{ r \\mid r \\in x_i, f(r) = 0 \\}`
    - |intercepts|

    Examples
    --------
    Import `intercept_points` function from `regressions` library
        >>> from regressions.analyses.intercepts import intercept_points
    Calculate the roots of a cubic function with coefficients 1, -15, 66, and -80
        >>> points_cubic = intercept_points('cubic', [1, -15, 66, -80])
        >>> print(points_cubic)
        [2.0, 5.0, 8.0]
    Calculate the roots of a sinusoidal function with coefficients 3, 1, -2, and 3
        >>> points_sinusoidal = intercept_points('sinusoidal', [3, 1, -2, 3])
        >>> print(points_sinusoidal)
        [-3.5708, 2.7124, 8.9956, 15.2788, 21.5619, '-3.5708 + 6.2832k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    positive_integer(precision)

    # Determine intercepts based on equation type
    result = []
    if equation_type == 'linear':
        result = linear_roots(*coefficients, precision)
    elif equation_type == 'quadratic':
        result = quadratic_roots(*coefficients, precision)
    elif equation_type == 'cubic':
        result = cubic_roots(*coefficients, precision)
    elif equation_type == 'hyperbolic':
        result = hyperbolic_roots(*coefficients, precision)
    elif equation_type == 'exponential':
        result = exponential_roots(*coefficients, precision)
    elif equation_type == 'logarithmic':
        result = logarithmic_roots(*coefficients, precision)
    elif equation_type == 'logistic':
        result = logistic_roots(*coefficients, precision)
    elif equation_type == 'sinusoidal':
        result = sinusoidal_roots(*coefficients, precision)
    return result


from math import log
from regressions.errors.scalars import two_scalars, three_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def exponential_roots(first_constant, second_constant, precision = 4):
    """
    Calculates the roots of an exponential function

    Parameters
    ----------
    first_constant : int or float
        Constant multiple of the original exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Base rate of variable of the original exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    roots : list of float
        List of the x-coordinates of all of the x-intercepts of the original function; if the function never crosses the x-axis, then it will return a list of `None`

    See Also
    --------
    :func:`~regressions.analyses.equations.exponential.exponential_equation`, :func:`~regressions.analyses.derivatives.exponential.exponential_derivatives`, :func:`~regressions.analyses.integrals.exponential.exponential_integral`, :func:`~regressions.models.exponential.exponential_model`

    Notes
    -----
    - Standard form of a exponential function: :math:`f(x) = a\\cdot{b^x}`
    - Exponential formula: :math:`x = \\varnothing`

    Examples
    --------
    Import `exponential_roots` function from `regressions` library
        >>> from regressions.analyses.roots.exponential import exponential_roots
    Calculate the roots of an exponential function with coefficients 2 and 3
        >>> roots_first = exponential_roots(2, 3)
        >>> print(roots_first)
        [None]
    Calculate the roots of an exponential function with coefficients -2 and 3
        >>> roots_second = exponential_roots(-2, 3)
        >>> print(roots_second)
        [None]
    Calculate the roots of an exponential function with all inputs set to 0
        >>> roots_zeroes = exponential_roots(0, 0)
        >>> print(roots_zeroes)
        [None]
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create list to return
    result = []
    
    # Determine root
    root = None

    # Return result
    result.append(root)
    return result

def exponential_roots_first_derivative(first_constant, second_constant, precision = 4):
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)

    # Create list to return
    result = []
    
    # Determine root of first derivative
    root = None

    # Return result
    result.append(root)
    return result

def exponential_roots_second_derivative(first_constant, second_constant, precision = 4):
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)

    # Create list to return
    result = []
    
    # Determine root of second derivative
    root = None

    # Return result
    result.append(root)
    return result

def exponential_roots_initial_value(first_constant, second_constant, initial_value, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, initial_value)
    positive_integer(precision)

    # Create list to return
    result = []

    # Create intermediary variables
    numerator = log(abs(initial_value / first_constant))
    denominator = log(abs(second_constant))

    # Circumvent division by zero
    if denominator == 0:
        denominator = 10**(-precision)

    # Determine root given an initial value
    ratio = numerator / denominator

    # Round root
    rounded_ratio = rounded_value(ratio, precision)

    # Return result
    result.append(rounded_ratio)
    return result

def exponential_roots_derivative_initial_value(first_constant, second_constant, initial_value, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, initial_value)
    positive_integer(precision)

    # Circumvent division by zero
    denominator = log(abs(second_constant))
    if denominator == 0:
        denominator = 10**(-precision)

    # Determine root of derivative given an initial value
    result = exponential_roots_initial_value(first_constant * denominator, second_constant, initial_value, precision)
    return result


from .positions import argument_position

def confirm_matrix(matrix, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(matrix, list) or not isinstance(matrix[0], list) or isinstance(matrix[0][0], list):
        raise TypeError(f'{identifier.capitalize()} must be a 2-dimensional list')
    else:
        return f'{identifier.capitalize()} is a 2-dimensional list'

def matrix_of_scalars(matrix, position = 'only'):
    confirm_matrix(matrix, position)
    identifier = argument_position(position)
    for vector in matrix:
        if not isinstance(vector, list):
            raise TypeError(f'Elements within {identifier} must be lists')
        for scalar in vector:
            if not isinstance(scalar, (int, float)):
                raise TypeError(f'Elements within lists within {identifier} must be integers or floats')
    else:
        return f'{identifier.capitalize()} is a 2-dimensional list containing nested elements that are integers or floats'

def square_matrix(matrix):
    matrix_of_scalars(matrix, 'first')
    if len(matrix) is not len(matrix[0]):
        raise ValueError('First argument must contain the same amount of lists as the amount of elements contained within its first list')
    else:
        return 'First argument contains the same amount of lists as the amount of elements contained within its first list'

def compare_rows(matrix_one, matrix_two):
    matrix_of_scalars(matrix_one, 'first')
    matrix_of_scalars(matrix_two, 'second')
    if len(matrix_one) is not len(matrix_two):
        raise ValueError('First argument and second argument must contain the same amount of lists')
    else:
        return 'First argument and second argument contain the same amount of lists'

def compare_columns(matrix_one, matrix_two):
    matrix_of_scalars(matrix_one, 'first')
    matrix_of_scalars(matrix_two, 'second')
    if len(matrix_one[0]) is not len(matrix_two[0]):
        raise ValueError('First list within first argument and first list within second argument must contain the same amount of elements')
    else:
        return 'First list within first argument and first list within second argument contain the same amount of elements'

def compare_matrices(matrix_one, matrix_two):
    compare_rows(matrix_one, matrix_two)
    compare_columns(matrix_one, matrix_two)
    return 'First argument and second argument contain the same amount of lists; first list within first argument and first list within second argument contain the same amount of elements'

def columns_rows(matrix_one, matrix_two):
    matrix_of_scalars(matrix_one, 'first')
    matrix_of_scalars(matrix_two, 'second')
    if len(matrix_one[0]) is not len(matrix_two):
        raise ValueError('First list within first argument must contain the same amount of elements as the amount of lists contained within second argument')
    else:
        return 'First list within first argument contains the same amount of elements as the amount of lists contained within second argument'

def allow_none_matrix(matrix, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(matrix, list):
        raise TypeError(f'{identifier.capitalize()} must be a list')
    if not isinstance(matrix[0], list) and matrix[0] is not None:
        raise TypeError(f'{identifier.capitalize()} must contain either lists or None')
    if len(matrix) > 1 and not isinstance(matrix[1], list):
        raise TypeError(f'Second element of {identifier} must be a list')
    if len(matrix) > 1 and not isinstance(matrix[0][0], (int, float)):
        raise TypeError(f'If the first element of {identifier} is not None, then it must be a list of integers or floats')
    if len(matrix) > 1 and not isinstance(matrix[-1][0], str):
        raise TypeError(f'If the first element of {identifier} is not None, then the last element must be a list with a first element that is a string')
    else:
        return f'{identifier.capitalize()} is either a 1-dimensional list that contains None or a 2-dimensional list containing lists in which the first nested list contains integers or floats and the last nested list begins with a string'

def allow_vector_matrix(matrix, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(matrix, list):
        raise TypeError(f'{identifier.capitalize()} must be a list')
    if not isinstance(matrix[0], (list, int, float, str)) and matrix[0] is not None:
        raise TypeError(f'Elements of {identifier} must lists, integers, floats, strings, or None')
    if len(matrix) > 1 and isinstance(matrix[0], list) and isinstance(matrix[0][0], list):
        raise TypeError(f'{identifier.capitalize()} cannot be more than a 2-dimensional list')
    else:
        return f'{identifier.capitalize()} is a list containing lists, integers, floats, strings, or None'

def level(matrix, scalar):
    if not scalar <= len(matrix[0]):
        raise ValueError('Last argument must be less than or equal to the length of the nested lists within the first argument')
    else:
        return 'Last argument is less than or equal to the length of the nested lists within the first argument'


from math import exp
from regressions.errors.scalars import two_scalars, three_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value
from .hyperbolic import hyperbolic_roots

def logarithmic_roots(first_constant, second_constant, precision = 4):
    """
    Calculates the roots of a logarithmic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the logarithmic term of the original logarithmic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the constant term of the original logarithmic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    roots : list of float
        List of the x-coordinates of all of the x-intercepts of the original function

    See Also
    --------
    :func:`~regressions.analyses.equations.logarithmic.logarithmic_equation`, :func:`~regressions.analyses.derivatives.logarithmic.logarithmic_derivatives`, :func:`~regressions.analyses.integrals.logarithmic.logarithmic_integral`, :func:`~regressions.models.logarithmic.logarithmic_model`

    Notes
    -----
    - Standard form of a logarithmic function: :math:`f(x) = a\\cdot{\\ln{x}} + b`
    - Logarithmic formula: :math:`x = \\text{e}^{-\\frac{b}{a}}`

    Examples
    --------
    Import `logarithmic_roots` function from `regressions` library
        >>> from regressions.analyses.roots.logarithmic import logarithmic_roots
    Calculate the roots of a logarithmic function with coefficients 2 and 3
        >>> roots_first = logarithmic_roots(2, 3)
        >>> print(roots_first)
        [0.2231]
    Calculate the roots of a logarithmic function with coefficients -2 and 3
        >>> roots_second = logarithmic_roots(-2, 3)
        >>> print(roots_second)
        [4.4817]
    Calculate the roots of a logarithmic function with all inputs set to 0
        >>> roots_zeroes = logarithmic_roots(0, 0)
        >>> print(roots_zeroes)
        [0.3679]
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create list to return
    result = []
    
    # Determine root
    root = exp(-1 * coefficients[1] / coefficients[0])

    # Round root
    rounded_root = rounded_value(root, precision)

    # Return result
    result.append(rounded_root)
    return result

def logarithmic_roots_first_derivative(first_constant, second_constant, precision = 4):
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)

    # Create list to return
    result = []
    
    # Determine root
    root = None

    # Return result
    result.append(root)
    return result

def logarithmic_roots_second_derivative(first_constant, second_constant, precision = 4):
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)

    # Create list to return
    result = []
    
    # Determine root
    root = None

    # Return result
    result.append(root)
    return result

def logarithmic_roots_initial_value(first_constant, second_constant, initial_value, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, initial_value)
    positive_integer(precision)

    # Determine roots given an initial value
    result = logarithmic_roots(first_constant, second_constant - initial_value, precision)
    return result

def logarithmic_roots_derivative_initial_value(first_constant, second_constant, initial_value, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, initial_value)
    positive_integer(precision)

    # Determine roots of derivative given an initial value
    result = hyperbolic_roots(first_constant, -1 * initial_value, precision)
    return result


from math import exp
from regressions.errors.scalars import three_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def logistic_equation(first_constant, second_constant, third_constant, precision = 4):
    """
    Generates a logistic function to provide evaluations at variable inputs

    Parameters
    ----------
    first_constant : int or float
        Carrying capacity of the resultant logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Growth rate of the resultant logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Value of the sigmoid's midpoint of the resultant logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First three arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    evaluation : func
        Function for evaluating a logistic equation when passed any integer or float argument

    See Also
    --------
    :func:`~regressions.analyses.derivatives.logistic.logistic_derivatives`, :func:`~regressions.analyses.integrals.logistic.logistic_integral`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.models.logistic.logistic_model`

    Notes
    -----
    - Standard form of a logistic function: :math:`f(x) = \\frac{a}{1 + \\text{e}^{-b\\cdot(x - c)}}`
    - |logistic_functions|

    Examples
    --------
    Import `logistic_equation` function from `regressions` library
        >>> from regressions.analyses.equations.logistic import logistic_equation
    Create a logistic function with coefficients 2, 3, and 5, then evaluate it at 10
        >>> evaluation_first = logistic_equation(2, 3, 5)
        >>> print(evaluation_first(10))
        2.0
    Create a logistic function with coefficients 100, 5, and 11, then evaluate it at 10
        >>> evaluation_second = logistic_equation(100, 5, 11)
        >>> print(evaluation_second(10))
        0.6693
    Create a logistic function with all inputs set to 0, then evaluate it at 10
        >>> evaluation_zero = logistic_equation(0, 0, 0)
        >>> print(evaluation_zero(10))
        0.0001
    """
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant], precision)

    # Create evaluation
    def logistic_evaluation(variable):
        evaluation = coefficients[0] * (1 + exp(-1 * coefficients[1] * (variable - coefficients[2])))**(-1)
        result = rounded_value(evaluation, precision)
        return result
    return logistic_evaluation


from regressions.errors.vectors import vector_of_scalars

def column_conversion(vector):
    """
    Converts a row vector into a column vector

    Parameters
    ----------
    vector : list of int or float
        List of numbers representing a vector

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list

    Returns
    -------
    column : list of lists of int or float
        List in which each element is a list containing an element from the input vector

    See Also
    --------
    :func:`~regressions.vectors.dimension.single_dimension`

    Notes
    -----
    - Row vector: :math:`\\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Column vector: :math:`\\left\\langle\\begin{matrix} a_1, \\\\ a_2, \\\\ \\cdots, \\\\ a_n \\end{matrix}\\right\\rangle`

    Examples
    --------
    Import `column_conversion` function from `regressions` library
        >>> from regressions.vectors.column import column_conversion
    Convert [1, 2, 3]
        >>> column_3d = column_conversion([1, 2, 3])
        >>> print(column_3d)
        [[1], [2], [3]]
    Convert [-7, 5]
        >>> column_2d = column_conversion([-7, 5])
        >>> print(column_2d)
        [[-7], [5]]
    """
    # Handle input errors
    vector_of_scalars(vector)

    # Create list to return
    result = []

    # Iterate over input
    for element in vector:
        # Store elements as lists within list to return
        result.append([element])
    
    # Return result
    return result


from regressions.errors.vectors import vector_of_strings

def sorted_list(data):
    """
    Sorts all elements in a data set in increasing order

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze

    Returns
    -------
    order : list of int or float
        List of all elements from a data set sorted in increasing order

    See Also
    --------
    :func:`~regressions.statistics.minimum.minimum_value`, :func:`~regressions.statistics.maximum.maximum_value`, :func:`~regressions.statistics.median.median_value`, :func:`~regressions.statistics.halve.half`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Sorted version of set: :math:`A_i = ( A_1, A_2, \\cdots, A_n )`

        - For all terms in :math:`A_i`: :math:`A_{n-1} \\leq A_n`

    - |monotonic_function|

    Examples
    --------
    Import `sorted_list` function from `regressions` library
        >>> from regressions.statistics.sort import sorted_list
    Sort the set [5, 2, 9, 8]
        >>> order_1 = sorted_list([5, 2, 9, 8])
        >>> print(order_1)
        [2, 5, 8, 9]
    Sort the set [11, 3, 52, 25, 21, 25, 6]
        >>> order_2 = sorted_list([11, 3, 52, 25, 21, 25, 6])
        >>> print(order_2)
        [3, 6, 11, 21, 25, 25, 52]
    """
    # Create intermediary lists
    pivots = []
    less = []
    more = []

    # Handle base case
    if len(data) <= 1:
        return data
    
    # Handle recursive case
    else:
        # Set pivot
        pivot = data[0]

        # Iterate over input
        for i in data:
            # Store values below pivot in less list
            if i < pivot:
                less.append(i)
            
            # Store values above pivot in more list
            elif i > pivot:
                more.append(i)
            
            # Store values equal to pivot in pivots list
            else:
                pivots.append(i)
        
        # Sort intermediary lists
        less = sorted_list(less)
        more = sorted_list(more)

        # Create final list to return
        result = less + pivots + more
        return result

def sorted_dimension(data, dimension = 1):
    """
    Sorts all elements in a multidimensional data set in increasing order, according to particular dimension

    Parameters
    ----------
    data : list of lists of int or float
        List of lists of numbers to analyze
    dimension: int, default=1
        Number representing the dimension to use for sorting

    Returns
    -------
    order : list of lists of int or float
        List of lists of numbers sorted in increasing order, based on only one dimension of the nested list

    See Also
    --------
    :func:`~regressions.statistics.halve.half_dimension`, :func:`~regressions.vectors.dimension.single_dimension`

    Notes
    -----
    - Set of ordered pairs of numbers: :math:`a_i = \\{ ( a_{1,1}, a_{1,2}, \\cdots, a_{1,j}, a_{1,n} ), ( a_{2,1}, a_{2,2}, \\cdots, a_{2,j}, a_{2,n} ), \\cdots, \\\\ ( a_{m,1}, a_{m,2}, \\cdots, a_{m,j}, a_{m,n} ) \\}`
    - Sorted version of set according to the values in the :math:`j`\ th position: :math:`A_i = ( ( A_{1,1}, A_{1,2}, \\cdots, A_{1,j}, A_{1,n} ), ( A_{2,1}, A_{2,2}, \\cdots, A_{2,j}, A_{2,n} ), \\cdots, \\\\ ( A_{m,1}, A_{m,2}, \\cdots, A_{m,j}, A_{m,n} ) )`

        - For all terms in :math:`A_i`: :math:`A_{n-1,j} \\leq A_{n,j}`

    - |monotonic_function|

    Examples
    --------
    Import `sorted_dimension` function from `regressions` library
        >>> from regressions.statistics.sort import sorted_dimension
    Sort the set [[1, 3, 5], [9, 2, 4], [6, 1, 8]] according to its second dimension
        >>> order_1 = sorted_dimension([[1, 3, 5], [9, 2, 4], [6, 1, 8]], 2)
        >>> print(order_1)
        [[6, 1, 8], [9, 2, 4], [1, 3, 5]]
    Sort the set [[1, 3, 5], [9, 2, 4], [6, 1, 8]] according to its third dimension
        >>> order_2 = sorted_dimension([[1, 3, 5], [9, 2, 4], [6, 1, 8]], 3)
        >>> print(order_2)
        [[9, 2, 4], [1, 3, 5], [6, 1, 8]]
    """
    # Create intermediary lists
    pivots = []
    less = []
    more = []

    # Handle base case
    if len(data) <= 1:
        return data
    
    # Handle recursive case
    else:
        # Set pivot
        pivot = data[0][dimension - 1]

        # Iterate over list
        for i in data:
            # Store values below pivot in less list
            if i[dimension - 1] < pivot:
                less.append(i)
            
            # Store values above pivot in more list
            elif i[dimension - 1] > pivot:
                more.append(i)
            
            # Store values equal to pivot in pivots list
            else:
                pivots.append(i)
        
        # Sort intermediary lists
        less = sorted_dimension(less, dimension)
        more = sorted_dimension(more, dimension)

        # Create final list to return
        result = less + pivots + more
        return result

def sorted_strings(data):
    # Handle input errors
    vector_of_strings(data)

    # Create list to return
    sorted_data = []

    # Handle single entry requiring no sorting
    if len(data) == 1:
        sorted_data = data
    
    # Handle general case of two entries
    else:
        # Determine values of strings
        first_index = data[0].find(' + ')
        first_value = float(data[0][:first_index])
        second_index = data[1].find(' + ')
        second_value = float(data[1][:second_index])
        
        # Set return list to initial list if already sorted
        if first_value < second_value:
            sorted_data = data
        
        # Set return list to inversion of initial list if not already sorted
        else:
            sorted_data = [data[1], data[0]]
    
    # Return result
    return sorted_data


def argument_position(position = 'only'):
    identifier = ''
    argument = 'argument'
    if position == 'only':
        identifier = argument
    else:
        identifier = position + ' ' + argument
    return identifier